/**
 * driver.h
 *
 * AMLET library
 * Copyright (C) 2002-2010 Fabian Bastin <bastin@iro.umontreal.ca>
 *
 * Licensed under the Open Software License version 2.1
 */

#ifndef __AMLET_LIBRARY_DRIVER_H__
#define __AMLET_LIBRARY_DRIVER_H__

#include <oratio/ctools.h>
#include <oratio/moca.h>
#include <amlet/dcl.h>

#include <amlet/model.h>

#ifdef __cplusplus
extern "C" {
#endif

  /* array identifiers */

  /* scalar identifiers */
  enum {
    NP,              /* number of people */
    NALT,            /* number of alternatives */
    NOBS,            /* number of observations */
    NVAR,
    IDDEP,           /* number of the variable that is the dependent
		        variable */
    CENSOR,
    IDCENSOR,

    RESCALE,         /* 1 if some independent variables has to be rescale,
			0 otherwise. */
    BIAS,            /* 1 is we apply the bias correction, 0 else. */

    NITER,           /* number of iterations allowed in the maximization
                        process */
    EPS,             /* tolerance for convergence */
    SEED1,           /* seed for random generator */
    NREP,            /* maximum sampling size */
    VARREP,          /* variable sample size indicator */
    LEVEL,           /* signification level for statistical tests */
    LEVEL_ACCURACY,  /* signification level for objective function
			evaluation */

    FORMAT,          /* data format */

    METHOD,          /* algorithm used for the optimization process */
    HESSIAN_UPDATE,  /* update strategy for the Hessian update */
    HESSIAN_APPROX,  /* use approximation instead of the true Hessian when
			computing statistical informations */

    RANGEN,          /* random numbers generator identifier */
    SAMPLING,        /* sampling method */
    RANDOMIZATION,   /* randomization method */
    QMCREP,          /* number of samplings for accuracy estimation with
			quasi-Monte Carlo draws */

    VALIDATION,      /* validation indicator */
    STATISTICS,      /* compute error and bias of simulation


		        (automatically done if MCBTRDA method is selected) */

    ROBUST,          /* robust estimation of the information matrix */

    /* sampling */
    DRAWS,
    ALTERNATIVES_SAMPLING,

    /* data information */
    SYNTHETIC,       /* true if we use synthetic data, false else */

    /* Nests */
    NNESTS,          /* number of nests */

    /* strings */
    NAME,

    /* filenames */
    DRAWSNAME,

    END_RESERVED_WORDS
  };

  /* Random and quasi-random numbers generators */

  static const c_symbol randomization_halton[] = {
    {"none", MOCA_HALTON_RANDOMIZATION_NONE},
    {"scrambling", MOCA_HALTON_RANDOMIZATION_SCRAMBLING},
    {"shifting", MOCA_HALTON_RANDOMIZATION_SHIFTING},
    {"shifting_scrambling", MOCA_HALTON_RANDOMIZATION_SHIFTING_SCRAMBLING},
    {"shuffling", MOCA_HALTON_RANDOMIZATION_SHUFFLING},
    {"shifting_shuffling", MOCA_HALTON_RANDOMIZATION_SHIFTING_SHUFFLING}
  };

  enum {
    BTYPE = 0,     /* type of parameters to estimate */
    BGENTYPE,
    B,
    BGEN,
    IGNORE,
    RESCALEIDX,
    RESCALEVAL,
    END_ARRAY_WORDS
  };

  enum {
    NESTS = 0,
    END_OTHERS
  };

  /* reserved words */
  static const c_symbol reserved_words[] = {
    {"np", NP},
    {"nalt", NALT},
    {"nobs", NOBS},
    {"nvar", NVAR},
    {"iddep", IDDEP},

    {"censor", CENSOR},
    {"idcensor", IDCENSOR},

    {"rescale", RESCALE},

    {"bias", BIAS},

    {"niter", NITER},
    {"eps", EPS},

    /* Random numbers and draws generator parameters */
    {"seed1", SEED1},
    {"nrep", NREP},
    {"varrep", VARREP},
    {"level", LEVEL},

    {"format", FORMAT},

    {"method", METHOD},
    {"hessian_update", HESSIAN_UPDATE},
    {"hessian_approx", HESSIAN_APPROX},

    {"rangen", RANGEN},
    {"sampling", SAMPLING},
    {"randomization", RANDOMIZATION},
    {"qmcrep", QMCREP},

    {"validation", VALIDATION},
    {"robust", ROBUST},
    {"statistics", STATISTICS},

    {"draws", DRAWS},
    {"alternatives_sampling", ALTERNATIVES_SAMPLING},

    {"synthetic", SYNTHETIC},

    {"nnests", NNESTS},

    {"problem", NAME},
    {"drawsname", DRAWSNAME},

    /* arrays */
    {"btype", BTYPE},
    {"bgentype", BGENTYPE},
    {"b", B},
    {"bgen", BGEN},
    {"ignore", IGNORE},
    {"rescalidx", RESCALEIDX},
    {"rescalval", RESCALEVAL},

    /* others */ 
    {"nests", NESTS}
  };

  /* double entry arrays */

  enum {
    ALTTYPES = 0,
    ALTVALUES,
    END_DBL_ARRAY_WORDS
  };

  /* arrays */
  static const c_symbol arrays[] = {
    {"alttypes", ALTTYPES},
    {"altvalues", ALTVALUES},
  };

  /* optimization methods */

  static const c_symbol methods[] = {
    {"none", NO_OPTIMIZATION},
    {"mcbtr", MCBTR},
    {"mcrtr", MCRTR},
    {"mcbtrda", MCBTRDA},
    {"mcrtrda", MCRTRDA},
    {"mcbfgs", MCBFGS},
    {"mcbfgsda", MCBFGSDA},
    {"mcbtrc", MCBTRC}
  };

  static const c_symbol hessian_updates[] = {
    {"bfgs", HESSIAN_BFGS},
    {"bhhh", HESSIAN_BHHH},
    {"sr1", HESSIAN_SR1}
  };

  static const c_symbol rangen[] = {
    {"ecuyer", RAN_LECUYER88B},
    {"standard", RAN_STANDARDMINIMAL},
    {"parkmiller", RAN_STANDARDMINIMAL},
    {"mt19937", RAN_MT19937}
  };

  typedef union {
    int    i;
    double d;
    char  *s;
  } driver_data;

  typedef struct {
    driver_data d;
    int type;
  } driver_var;

  typedef struct {
    driver_var *array;
    int n;
  } driver_array;


#define RESERVED_WORDS NP
#define ARRAY_WORDS    B

  typedef struct {
    driver_data var[END_RESERVED_WORDS];
    driver_array array[END_ARRAY_WORDS];

    /* TODO: put them in the reserved words */
    int generation;
    char *obsname; /* name of the observations file */
    char *output;  /* name of the output file */

    driver_array *nests;   /* nests gestion */

    /* for the generation */
    int **alt_types;
    double **alt_values;
  } driver;

  /**
   * Assign the read values to the model.
   *
   * @param dr is the built driver.
   * @param m is an initialized AMLET model.
   */
  void driver_assign_model(driver *dr, AMLET_Model *m);

#ifdef __cplusplus
}
#endif

#endif  /* __AMLET_LIBRARY_DRIVER_H__ */
